#  Copyright (c) 2021. Taylor-Jayde Blackstone <t.blackstone@inspyre.tech> https://inspyre.tech
import os
import ctypes
import psutil
from datetime import datetime
from pypattyrn.behavioral.null import Null

from inspyre_toolbox.proc_man.errors import NoFoundProcessesError
from inspyre_toolbox.core_helpers.logging import add_isl_child, force_lowkey_log_name
from inspyre_toolbox.humanize import Numerical
from inspy_logger import InspyLogger

import subprocess


fts = datetime.fromtimestamp


ISL = InspyLogger('InspyreToolBox.ProcMan', 'debug')

LOG = ISL.device.start()


class Colors(object):
    def __init__(self, return_null=True):
        if return_null:
            self.red = Null()
            self.yellow = Null()
            self.blue = Null()
            self.green = Null()
            self.end_color = Null()
        else:
            from inspyred_print import Color, Format
            self.red = Color.red
            self.yellow = Color.yellow
            self.blue = Color.blue
            self.green = Color.green
            self.end_color = Format.end_mod


def isAdmin():
    try:
        is_admin = (os.getuid() == 0)
    except AttributeError:
        is_admin = ctypes.windll.shell32.IsUserAnAdmin() != 0
    return is_admin


def find_all_by_name(name, case_sensitive=False, inspy_logger_device=None, on_the_dl=False, colorful_logging=False):
    """

    Find a running program by name.

    Args:
        name (str):
            The name of the program you're looking for.

        case_sensitive (bool):
            Should the function care about letter casing when searching for the given program? (
            Optional) - Defaults to bool(False)

        inspy_logger_device (inspy_logger.InspyLogger().device):
            The instantiated device object from the inspy_logger package. (Optional, defaults to None)

            Note:
                If a value other than NoneType is passed to this parameter, the function will check the
                InspyLogger.device attribute 'started' to see if the logger has already been started elsewhere. One
                of three situations could result from this:

                    1) The attribute 'started' exists, and evaluates to bool(False):

                       In this case the function will attempt to start the log-device by calling it's 'start' method
                       before calling 'add_child' on the log-device in the hopes of receiving it's own logging object.

                    2) The attribute 'started' exists and evaluates to bool(False):

                       In this case the function will attempt to call 'add_child' on the log-device in the hopes of
                       receiving it's own logging object.

                    3) The attribute 'started' does not exist:

                       In this case, the function will receive a raised AttributeError from the passed object. If
                       this occurs; the function will raise a more relevant 'InvalidLogDeviceError'

        on_the_dl (bool):
            Should InspyreToolbox leave itself out of the the logging name? (Optional, Defaults to bool(False))

        colorful_logging (bool):
            Should the function log colorfully? (Optional, Defaults to bool(False))


    Returns:
        procs_found (list): A list of the currently running processes that match the given name.

    """

    isl_dev = inspy_logger_device

    prefix = '' if on_the_dl else 'InspyreToolbox.'
    log_name = prefix + 'proc_man.find_by_name'

    log = add_isl_child(log_name, isl_dev)

    if not force_lowkey_log_name:
        r_null = not colorful_logging
    else:
        r_null = force_lowkey_log_name

    colors = Colors(return_null=not colorful_logging)

    procs_found = []

    # If we're not being case-sensitive we'll lower the case on 'name'
    if not case_sensitive:
        log.debug('Searching in case insensitive mode!')
        name = name.lower()

    log.debug(f'Search query: {name}')

    # Iterate through the process list
    for proc in psutil.process_iter():
        try:
            # Gather the information of all the currently running processes
            proc_info = proc.as_dict(
                attrs=['pid', 'name', 'create_time', 'username'])
            log.debug(proc_info)

            # If we're not caring about case let's go ahead and make both
            # things we're comparing lowercase.
            if not case_sensitive:
                proc_name = proc_info['name'].lower()
            else:
                proc_name = proc_info['name']

            if name in proc_name:
                log.debug(f'{colors.green}Found match: {colors.blue}{proc_info["name"]}'
                          f'({colors.yellow}{proc_info["pid"]}{colors.blue})')

                # Add find to the list we're returning to the caller.
                procs_found.append(proc_info)

                log.debug(
                    f'{colors.green}Added {proc_info["pid"]} to found process list')
        except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
            pass

    num_found = Numerical(len(procs_found), noun='process')

    log.debug(f'Found: {colors.yellow}{num_found.count_noun()}')

    # Return found process list to caller
    return procs_found


def kill_all_in_list(kill_list, inspy_logger_device=None, on_the_dl=False, colorful_logging=False):
    """

    When provided a list of processes kill each process in the list.

    Args:
        kill_list:
            A list of processes in the format of a dictionary
        inspy_logger_device:
            An instantiated inspy-logger device.
        on_the_dl:
            Don't include 'InspyreToolbox' in the name of the logger.
        colorful_logging:
            Should logging be colorful?

    Returns:
        None

    """

    prefix = '' if on_the_dl else 'InspyreToolbox.'
    log_name = prefix + 'ProcMan.kill_all_in_list'

    log = add_isl_child(log_name, ISL.device)

    log.debug('Logging active')

    colors = Colors(return_null=not colorful_logging)

    missing = 0

    for proc in kill_list:
        log.debug(f'Killing {colors.yellow}{proc["pid"]}')
        os.kill(proc['pid'], 2)
        # try:
        #         subprocess.check_output("Taskkill /PID %d /F" % proc['pid'])
        #     if os.name == 'nt':
        #     else:
        #         os.kill(proc['pid'], 2)
        # except subprocess.CalledProcessError:
        #     missing += 1
        #     if missing >= 20:
        #         kill_all_by_name(proc['name'])
        #         break
        # log.debug(
        #     f'{colors.green}Kill signal sent to {colors.yellow}{proc["pid"]}')


def kill_all_by_name(name, case_sensitive=False, inspy_logger_device=None, on_the_dl=False, colorful_logging=False):
    """

    Kill all currently running programs (and all instances thereof) which have a name containing the string given as a
    value to the 'name' parameter.

    Note:
        Unless you are running this as administrator, any programs that you don't have permission to kill will not be
        killed.

        IF YOU ARE RUNNING AS ADMINISTRATOR EVERY PROCESS THAT MATCHES WILL BE KILLED REGARDLESS OF WHO INSTANTIATED
        THE PROCESS!!!!

    Args:
        name (str): The name of the program you'd like to kill all instances of. Whatever you provide here will be
        used as a query to find all programs that have name that contains this string.

        case_sensitive (bool): Tell the function that the letter casing is irrelevant to finding matching programs. (
        Optional, defaults to False)

        inspy_logger_device (inspy_logger.InspyLogger().device):
            The instantiated device object from the inspy_logger package. (Optional, defaults to None)

            Note:
                If a value other than NoneType is passed to this parameter, the function will check the
                InspyLogger.device attribute 'started' to see if the logger has already been started elsewhere. One
                of three situations could result from this:

                    1) The attribute 'started' exists, and evaluates to bool(False):

                       In this case the function will attempt to start the log-device by calling it's 'start' method
                       before calling 'add_child' on the log-device in the hopes of receiving it's own logging object.

                    2) The attribute 'started' exists and evaluates to bool(False):

                       In this case the function will attempt to call 'add_child' on the log-device in the hopes of
                       receiving it's own logging object.

                    3) The attribute 'started' does not exist:

                       In this case, the function will receive a raised AttributeError from the passed object. If
                       this occurs; the function will raise a more relevant 'InvalidLogDeviceError'

        on_the_dl (bool):
            Should InspyreToolbox leave itself out of the the logging name? (Optional, Defaults to bool(False))

        colorful_logging (bool):
            Should the function log colorfully? (Optional, Defaults to bool(False))

    Returns:
        None

    """
    from time import sleep

    # Get a logger for logging in logging is enabled.
    log = add_isl_child('proc_man.kill_all_by_name', inspy_logger_device)

    # Get colors for colorful logging if colorful_logging is bool(True)
    color = Colors(return_null=not colorful_logging)

    log.debug(f'{color.yellow}Checking to see if there are, any running programs with "'
              f'{color.green}{name}{color.yellow}" in their name.')

    # Get a list of all the currently running processes that contains 'name' string.
    procs = find_all_by_name(name, case_sensitive,
                             inspy_logger_device, on_the_dl, colorful_logging)

    log.debug(procs)

    sleep(5)

    # Get the number of running processes found.
    num_running = Numerical(len(procs), noun='process')

    log.debug(f'Found {num_running.count_noun()}')

    # Gather the username of the current user.
    username = os.getlogin()

    # Find out if they are an administrator or not.
    is_admin = isAdmin()

    started_by_user = []

    if not is_admin:
        for proc in procs:
            log.debug(proc['username'])
            if proc['username'].endswith(username):
                started_by_user.append(proc)
        
    started_by_user.reverse()
    
    kill_all_in_list(started_by_user)



#  Copyright (c) 2021. Taylor-Jayde Blackstone <t.blackstone@inspyre.tech> https://inspyre.tech
from pypattyrn.behavioral.null import Null


force_lowkey_log_name = False
"""

This variable should be a boolean indicating whether or not InspyreToolbox should leave itself out when it's
functions name their child loggers.

Note:


For Example:

    Let's look at how 'inspyre_toolbox.proc_man.find_all_by_name' would name it's loggers in each of the situations we
    could find ourselves in;
   
    Note:
        [ROOT_LOGGER_NAME] is just a placeholder for whatever the name of the root logging device is.

    Case 1 (lowkey_log_name evaluates to bool(False)):
        
        [ROOT_LOGGER_NAME].InspyreToolbox.proc_man.find_all_by_name
        
    Case 2 (lowkey_log_name evaluates to any value seen as bool(True)):
        
        [ROOT_LOGGER_NAME].proc_man.find_all_by_name
        

"""


class InvalidLogDeviceError(Exception):

    def __init__(self):
        """

        Raised when one of the functions in InspyreToolbox are given a value to "inspy_logger_device" and it's found
        that the object provided is not properly configured, thus receiving an AttributeError when trying to check
        inspy_logger_device.started for a bool() value.

        Note:
            'inspy_logger_device.started' will exist in all cases where the logger device was given a name and
            log-level on instantiation (or later through setter methods). If the logging device hasn't been set up,
            an AttributeError will cause the following functions to raise InvalidLogDeviceError:

                * find_by_name
                * kill_all_by_name

        """
        self.message = 'The passed device does not appear to have been initialized, or is not an InspyLogger device!'


def add_isl_child(name, isl_device):
    """

    Manage using an InspyLogger device with Inspyre Toolbox.

    Args:
        name:
            The name you'd like for your log-device.

        isl_device (inspy_logger.InspyLogger.device | None):
            An instantiated inspy-logger device.

    Returns:
        The log device.

    """
    if isl_device:
        try:
            if not isl_device.started:
                log = isl_device.start()
        except AttributeError:
            raise InvalidLogDeviceError()

        log = isl_device.add_child(name)
    else:
        log = Null()

    return log



"""
File: inspyre_toolbox/live_timer/__init__.py
Created: ??

Description:
    inspyre_toolbox.live_timer gives you access to a class and functions that make running a live
    timer that's not only easily to initialize, but easy to query, reset, and pause your timers.
    This module also contains a class named 'TimerHistory' that keeps a history of all timer actions.
    
    To find out more about usage please see:
    
    GIT_REPO_ROOT/examples/live_timer

"""

from os import makedirs
from pathlib import Path
from time import time


class TimerHistory(object):

    def __init__(self, elapsed_method):
        self.get_elapsed = elapsed_method
        self.ledger = []
        self.actions = [
                "START",
                "STOP",
                "PAUSE",
                "UNPAUSE",
                "RESET",
                "CREATE",
                "QUERY"
                ]
        self.add("CREATE")

    def add(self, action: str = "START"):
        action = action.upper()
        entry = {
            "time": time(),
            "elapsed_since_last": "",
            "action": action,
            "rt_at_create": ""
        }
        if action == "CREATE":
            entry['elapsed_since_last'] = 0.00
        else:
            entry['elapsed_since_last'] = self.get_elapsed(self.ledger[-1]['time'])
            entry['rt_at_create'] = self.get_elapsed(self.ledger[0]['time'])

        self.ledger.append(entry)

    def write(self):
        data_path = Path("~/Inspyre-Softworks/Inspyre-Toolbox/data").expanduser()

        filename = "ledger_" + str(time()).split('.')[0]
        filepath = str(str(data_path) + "/" + filename + ".txt")

        filepath = str(Path(filepath).resolve())

        if not data_path.exists():
            makedirs(data_path)

        with open(filepath, "w") as fp:
            fp.write(str(self.ledger))


def format_seconds_to_hhmmss(seconds):
    hours = seconds // (60 * 60)
    seconds %= (60 * 60)
    minutes = seconds // 60
    seconds %= 60
    return "%02i:%02i:%02i" % (hours, minutes, seconds)


class Timer(object):
    
    def __repr__(self):
        statement = "Timer("\
                    f"Started: {self.started} |"
        if self.started:
            runtime = time() - self.start_time
            runtime = format_seconds_to_hhmmss(runtime)
            statement += f" Started: {self.start_time} - Current Runtime: {runtime}"

    def __init__(self):
        
        # Define some default attribute values
        
        self.start_time = None
        self.is_running = False
        self.pause_start = time()
        self.pause_end = None
        self.total_pause_time = 0
        self.started = False
        self.paused = False
        self.was_paused = False
        self.mark_2 = None

        # Start a Timer history object to track times for resets
        self.history = TimerHistory(self.get_elapsed)

    def get_elapsed(self, ts=None, sans_pause: bool = False, seconds=False):
        """
        
        Args:
            ts:
            sans_pause:
            seconds:

        Returns:

        """
        if ts is None:
            diff_time = self.start_time
        else:
            diff_time = ts

        self.mark_2 = time()
        # print(self.mark_2)
        # print(self.start_time)

        diff = self.mark_2 - diff_time
        # print(format_seconds_to_hhmmss(diff))
        if sans_pause:
            return format_seconds_to_hhmmss(diff)

        tpt = 0
        # print(self.total_pause_time)
        if self.paused:
            tpt += time() - self.pause_start

        tpt += self.total_pause_time
        diff = diff - tpt

        if seconds:
            return diff
        else:
            return format_seconds_to_hhmmss(diff)

    def reset(self):
        """

        Reset the 'self.start_time' attribute to this moment, effectively resetting the timer to '00:00:00'

        """
        self.history.add(action="RESET")
        self.start()

    def start(self):
        """

        Store the time the thread was started and assign the attribute 'self.started' to 'True' to indicate this.

        """
        self.start_time = time()
        self.started = True
        self.history.add()

    def pause(self):
        """

        Pause the running timer.

        This function will pause the running timer. What this really means in this context is that this function will
        fill the 'self.pause_start' time with the current time. When you unpause at a later time the 'self.unpause'
        function will reference 'self.pause_start' for comparison.

        This function also sets the 'self.paused' variable to 'True'.

        :return:
        """
        if not self.paused:
            self.pause_start = time()
            self.paused = True
            self.history.add("PAUSE")
        else:
            return False

    def unpause(self):
        """

        Un-Pause the running timer.

        This function will unpause the running timer. What that really means in this context is that since it marks the
        time that one paused the timer previously and this will compare the time elapsed between when the pause function
        was called and when this is called and keeps track of it.

        Whenever get_elapsed is called, it looks at the amount of seconds that have been added by this function and
        removes that total from the number of seconds elapsed before passing them to be naturalized.

        :return:
              bool
        """
        if self.paused:
            self.pause_end = time()
            diff = self.pause_end - self.pause_start
            self.total_pause_time += diff
            self.paused = False
            self.history.add("UNPAUSE")
        else:
            return False



#  Copyright (c) 2021. Taylor-Jayde Blackstone <t.blackstone@inspyre.tech> https://inspyre.tech
import os

def add_to_path(program_path:str):
    """Takes in a path to a program and adds it to the system path"""
    if os.name == "nt": # Windows systems
        import winreg # Allows access to the windows registry
        import ctypes # Allows interface with low-level C API's

        with winreg.ConnectRegistry(None, winreg.HKEY_CURRENT_USER) as root: # Get the current user registry
            with winreg.OpenKey(root, "Environment", 0, winreg.KEY_ALL_ACCESS) as key: # Go to the environment key
                existing_path_value = winreg.EnumValue(key, 3)[1] # Grab the current path value
                new_path_value = existing_path_value + program_path + ";" # Takes the current path value and appends the new program path
                winreg.SetValueEx(key, "PATH", 0, winreg.REG_EXPAND_SZ, new_path_value) # Updated the path with the updated path

            # Tell other processes to update their environment
            HWND_BROADCAST = 0xFFFF
            WM_SETTINGCHANGE = 0x1A
            SMTO_ABORTIFHUNG = 0x0002
            result = ctypes.c_long()
            SendMessageTimeoutW = ctypes.windll.user32.SendMessageTimeoutW
            SendMessageTimeoutW(HWND_BROADCAST, WM_SETTINGCHANGE, 0, u"Environment", SMTO_ABORTIFHUNG, 5000, ctypes.byref(result),)
    else: # If system is *nix
        with open(f"{os.getenv('HOME')}/.bashrc", "a") as bash_file:  # Open bashrc file
            bash_file.write(f'\nexport PATH="{program_path}:$PATH"\n')  # Add program path to Path variable
        os.system(f". {os.getenv('HOME')}/.bashrc")  # Update bash source
    print(f"Added {program_path} to path, please restart shell for changes to take effect")




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun  9 01:23:36 2021

@author: taylor
"""
from inflect import engine
import typing

from inspyre_toolbox.humanize.errors.numerical import NumericalErrors

# Instantiate inflect.engine as 'INF'
INF = engine()

NUM_ERR = NumericalErrors()


class NumericalStrings(object):
    """

    Number statement string fragments

    Attributes:
        less_than_20 (list): A list of strings each representing a number from 0 (or "") to 19 (or Nineteen).
        tens (list): A list of strings each representing a 'tens' place with all strings below ten being represented
        by "" and with the final element being "Ninety"
        thousands (list): A list of strings each representing a 'thousands' place with all strings below a thousand
        being represented by "" and with the final element being "Billion"
    """

    less_than_20 = [
        "",
        "One",
        "Two",
        "Three",
        "Four",
        "Five",
        "Six",
        "Seven",
        "Eight",
        "Nine",
        "Ten",
        "Eleven",
        "Twelve",
        "Thirteen",
        "Fourteen",
        "Fifteen",
        "Sixteen",
        "Seventeen",
        "Eighteen",
        "Nineteen",
    ]
    tens = [
        "",
        "Ten",
        "Twenty",
        "Thirty",
        "Forty",
        "Fifty",
        "Sixty",
        "Seventy",
        "Eighty",
        "Ninety",
    ]
    thousands = ["", "Thousand", "Million", "Billion"]


class Numerical(object):
    """

    A class that allows you to manipulate numbers with greater-flexibility.

    """
    
    def __add__(self, other_num, force_return_count=False, force_return_self=False):
        self.number = self.number + other_num

        if force_return_count and self.noun is None:
            raise ValueError()
        if force_return_count:
            return self.count_noun()
        elif force_return_self:
            return self
        else:
            return self.number
        
    def __mul__(self, multiplied_by, force_return_count=False, force_return_self=False):
        self.number = self.number * float(multiplied_by)

        if force_return_count and self.noun is None:
            raise ValueError()
        if force_return_count:
            return self.count_noun()
        elif force_return_self:
            return self
        else:
            return self.number
    
    def __rtruediv__(self, principle, force_return_count=False, force_return_self=False):
        if principle == 0 or self.number == 0 :
            raise ZeroDivisionError()

        self.number = float(principle) / self.number

        if force_return_count and self.noun is None:
            raise ValueError()
        if force_return_count :
            return self.count_noun()
        elif force_return_self :
            return self
        else :
            return self.number
    
    def __truediv__(self, divide_by, force_return_count=False, force_return_self=False):
        if divide_by == 0 or self.number == 0:
            raise ZeroDivisionError()

        self.number = self.number / float(divide_by)

        if force_return_count and self.noun is None:
            raise ValueError()
        if force_return_count:
            return self.count_noun()
        elif force_return_self:
            return self
        else:
            return self.number
            
    def __sub__(self, other, force_return_count=False, force_return_self=False):
        self.number = self.number - float(other)

        if force_return_count and self.noun is None:
            raise ValueError()
        if force_return_count:
            return self.count_noun()
        elif force_return_self:
            return self
        else:
            return self.number
    
    
    def __init__(self, number, noun=None, store_as_float=True):
        """

        Create a new instance, with a new number to manipulate

        Args:
            number (int|float): The subject number.
            noun: The noun to count.
        """
        # First, check whether 'number' is an integer or a float or not. If not; we raise an error
        
        self.store_as_float = store_as_float
        
        if not isinstance(number, float) and store_as_float:
            try:
                number = int(number)
            except ValueError as e:
                raise NUM_ERR.NotANumberError(
                    var=number,
                    var_name="number",
                    msg=f"Type {type(number)} is not valid. Must be {type(1)} or {type(1.1)}",
                )
            
            if store_as_float:
                number = float(number)

        self.number = number

        self.num_commified = self.commify(self.number)
        self.noun = noun

    def count_noun(
            self,
            number = None,
            noun: str = None,
            save_number: bool = True,
            save_noun: bool = True,

            **kwargs):
        """
        
        Args:
            number (Optional, (int|float)):
                The number of nouns there are. Defaults to 'self.number'.
            
            noun (Optional, str):
                The thing you want to describe the number of. Defaults to 'self.noun'.
            
            save_number (Optional, bool):
                If you provided a value for the 'number' parameter this parameter's value will indicate whether the
                value of 'number' should replace 'self.number'. Defaults to bool(True).
            
            save_noun (Optional, bool):
                If you provided a value for the 'noun' parameter and self.noun isn't NoneType and differs from the
                provided value 'save_noun' will indicate whether the new value for 'noun' should replace 'self.noun'. Defaults to bool(True).
            
            **kwargs:
            
                full_stop (bool):
                 Include a full-stop punctuation mark at the end of the final statement. Users that
                 are from America may know this punctuation mark better as 'period' (or; '.'. See
                 below for the 'period' parameter alias)
                 (Defaults to boo  l(False)

                period (bool):
                    Include a period mark at the end of the final statement. Users from some areas may
                    know this mark better as 'full-stop'
    
                noun (str):
                    Returns a string with the appropriate statement iy,.hy,i for annotating counts of various nouns.
    
                count (int/float, optional):
                    The number of 'noun' that exists.
                    (Defaults to 'Numerical.number')
    
                only_noun (bool, optional):
                    Return only the properly pluralized noun.
                    (Defaults to False)
    
                skip_commify (bool, optional):
                    (Ignored if either 'only_noun' or 'count_to_words' are True) Return the number as
                    part of the statement with commas added to the proper places.
                    (Defaults to False)
    
                to_words (bool, optional):
                    (Ignored if 'only_noun' is True) Modify the first part of the statement
                    (which contains the count) to words instead of numbers.
                    (Defaults to False.)
    
                capitalize (bool, optional):
                    Return the words for the count with the first letter capitalized.
                    (Defaults to False, ignored if 'count_to_words' is False.)
    
                round_num (int, optional):
                    Round the number returned to a given precision in decimal digits.
                    (Defaults to None(NoneType), ignored if 'only_noun' is True.)

        Returns:

        """

        number_changed = False
        noun_changed = False
        number_previous = None
        noun_previous = None

        if number is not None:
            try:
                number = float(number)
            except ValueError:

                warning = "If passing a noun alone to 'count_noun' you need to include the parameter name\n"\
                      "For example:\n"\
                      "\n"\
                      "number.count_noun(noun='banana')\n"\
                      "NOT\n"\
                      "number.count_noun('banana')"

                print(warning)

                raise ValueError(warning)

            if number != self.number:
                if not save_number:

                    number_changed = True
                    number_previous = self.number

                self.number = number


        if noun is not None:

            noun = str(noun)
            try:
                noun = str(noun)
                if self.noun is not None and noun != self.noun:
                    if not save_noun:
                        noun_changed = True
                        noun_previous = self.noun

                    self.noun = noun
            except:
                raise ValueError(f"'noun' should be {type(str())} not {type(noun)}")



        ret = self.__count_noun(noun=self.noun, **kwargs)

        if number_changed:
            self.number = number_previous
            number_previous = None
            number_changed = False
        if noun_changed:
            self.noun = noun_previous
            noun_previous = None
            noun_changed = None

        return ret

    def __count_noun(
        self,
        noun: str,
        count: int = None,
        only_noun=False,
        skip_commify=False,
        to_words=False,
        capitalize=False,
        full_stop=False,
        period=False,
        round_num=None
    ):
        """

        An alias for inflect.engine().plural_noun

        Returns the appropriate string for annotating counts of a given noun.

        Parameters:
            full_stop (bool):
                 Include a full-stop punctuation mark at the end of the final statement. Users that
                 are from America may know this punctuation mark better as 'period' (or; '.'. See
                 below for the 'period' parameter alias)
                 (Defaults to bool(False)

            period (bool):
                Include a period mark at the end of the final statement. Users from some areas may
                know this mark better as 'full-stop'

            noun (str):
                Returns a string with the appropriate statement iy,.hy,i for annotating counts of various nouns.

            count (int/float, optional):
                The number of 'noun' that exists.
                (Defaults to 'Numerical.number')

            only_noun (bool, optional):
                Return only the properly pluralized noun.
                (Defaults to False)

            skip_commify (bool, optional):
                (Ignored if either 'only_noun' or 'count_to_words' are True) Return the number as
                part of the statement with commas added to the proper places.
                (Defaults to False)

            to_words (bool, optional):
                (Ignored if 'only_noun' is True) Modify the first part of the statement
                (which contains the count) to words instead of numbers.
                (Defaults to False.)

            capitalize (bool, optional):
                Return the words for the count with the first letter capitalized.
                (Defaults to False, ignored if 'count_to_words' is False.)

            round_num (int, optional):
                Round the number returned to a given precision in decimal digits.
                (Defaults to None(NoneType), ignored if 'only_noun' is True.)

        Raises:

            ValueError: Raised when a value is provided for 'noun' that is not a string.

        """

        # Make sure our noun is a string or raise ValueError
        if not isinstance(noun, str):
            raise ValueError("Noun must be of type: str")

        # Make sure 'count' is not 'None' or we can't do anything with it.
        #
        # If 'count' is indeed 'None' we'll use the number that seeded this instance of the
        # 'Numerical' class.
        if count is None:
            count = self.number

        # If we received the parameter to round the results, we'll do that.
        if round_num is not None and isinstance(round_num, int):
            count = round(count, round_num)

        # Pluralize the noun string
        n_noun = INF.plural_noun(noun, count)

        # If the parameter 'to_words' is true, we need to convert the number to words (using the
        # 'to_words' function of this class before we finally concatenate our results.
        if to_words:
            c_count = self.to_words(count)
        else:

            # If we did not receive a bool(True) value for the 'skip_commify' parameter we'll
            # send our number off to be commified by self.commify before concatenation
            c_count = self.commify(target_number=count) if not skip_commify else count
        # If the parameter 'only_noun' evaluates to bool(True) the request should not contain
        # the root number in any form.
        statement = n_noun if only_noun else f"{c_count} {n_noun}"
        # If the 'capitalize' parameter evaluates to bool(True) we capitalize the final statement, as
        # one capitalizes a sentence.8h
        if capitalize:
            statement = statement.capitalize()

        # If the parameter 'full_stop' evaluates to bool(True) a full-stop (or; 'period' to most of my
        # fellow Americans.
        if full_stop or period:
            statement += "."

        # Return the appropriate pluaralized (or not) noun based off the number provided
        return statement

    def to_str(self):
        """

        Return 'self.number' as a string type without modifying it at  all from it's original form.

        Returns:
            (int/float): The contents found at 'self.number' but in as a string type.

        """
        return str(self.number)

    def commify(self, target_number: int = None):
        """

        Return a string containing your number as a string with commas added

        Parameters:
            target_number (int, optional): The number you'd like returned as a commified string. Defaults to
            'Numerical.number'

        Returns:
            str: self.number converted to a string and with appropriately placed commas.

        """
        num = self.number if target_number is None else target_number
        return "{:,}".format(num)

    def to_words(self, target_num=None):
        """

        An alias to inflect.engine().number_to_words()

        Literally just passes 'target_num' to INF.number_to_words and returns the result of that

        Parameters:
            target_num (int, optional): The number you'd like returned in word-form. Defaults to 'Numerical.number'.

        Returns:
            string: A string representing your target number in words.

        Raises:
            NotANumberError: Raised if 'target_num' isn't an integer or a float.

        """

        if target_num is None:
            target_num = self.number

        if not isinstance(target_num, int) and not isinstance(target_num, float):
            raise ValueError(
                "The parameter 'target_num' needs to be an integer or a float"
            )

        return INF.number_to_words(target_num)

    def __repr__(self):
        return self.to_words()



#  Copyright (c) 2021. Taylor-Jayde Blackstone <t.blackstone@inspyre.tech> https://inspyre.tech
"""

A package containing decorators that are a quick help in programming with Python

"""
from contextlib import contextmanager
from os import devnull
import sys

@contextmanager
def suppress_stdout():
    with open(devnull, "w") as dn:
        old_stdout = sys.stdout
        sys.stdout = dn
        try:
            yield
        finally:
            sys.stdout = old_stdout




#  Copyright (c) 2021. Taylor-Jayde Blackstone <t.blackstone@inspyre.tech> https://inspyre.tech
#  Added in v1.1


class NumericalErrors(object):
    def __init__(self):
        self.NotANumberError = NotANumberError


class NotANumberError(Exception):
    
    def __init__(self, var, var_name: str = None, msg: str = None, skip_print: bool = False, redirect_print=None) :
        """

        Raised when a parameter is provided that is not an integer or a float where one was expectecd.

        Args:
            var (Any): The variable which triggered the raising of this exception.
            var_name (str, optional): The name of the variable that was passed to raise this exception. Defaults to
            None.
            msg (str, optional): The message you'd like the end-user to see upon raising. Defaults to None.
            skip_print (bool, optional): Should we skip printing the message on initialization. Defaults to False.
            redirect_print (any output, optional): An object to redirect our printing to. Ignored if 'skip_print' is
            True. Defaults to None.


        Returns:
            None.

        """
        
        if var_name is None :
            msg1 = "The variable provided is neither an integer or a float. One or the other is needed."
        else :
            var_type = type(var)
            msg1 = f"{var_name} has a value of {var} and is of type {var_type}. An integer or float was needed."
        
        if msg is None :
            msg = msg1
        else :
            msg = msg1 + ' ' + msg
        
        self.msg = msg
        self.message = self.msg
        
        if not skip_print :
            if redirect_print :
                redirect_print(self.message)
            else :
                print(self.message)



from configparser import ConfigParser
from pkgutil import get_data

from inspyred_print import Color, Format

from inspyre_toolbox.pypi.packages import up_to_date

_color = Color()
_format = Format()
_red = _color.red
_end = _format.end_mod

__ver_parser = ConfigParser()
__ver = get_data('inspyre_toolbox', 'VERSION').decode('UTF8')
__ver_parser.read_string(__ver)

print(__ver_parser.sections())


class Version(object):

    def _check_most_current(self):
        res = up_to_date('inspy_logger')

        self.is_latest = res

        return res

    def __init__(self, version_info):
        print(dir(version_info))
        self.info = version_info['VERSION']
        self.number = self.info['number']
        self.pre_release = version_info.getboolean('VERSION', 'pre-release')

        if self.pre_release:
            self.pr_info = version_info['PRE-RELEASE']
            self.pr_type = self.pr_info['type']
            self.pr_num = self.pr_info['build-no']
            if self.pr_type.lower() == 'alpha':
                prt_tag = 'a'
            elif self.pr_type.lower() == 'beta':
                prt_tag = 'b'
            self.pr_tag = str(f"{prt_tag}{self.pr_num}")
        else:
            self.pr_info = None
            self.pr_type = None
            self.pr_num = None
            self.pr_tag = ""

        self.full = str(f"{self.number}{self.pr_tag}")
        self.is_latest = None
        self._check_most_current()


VERSION = Version(__ver_parser)
UPDATE_PENDING = not VERSION.is_latest
